# coding: utf-8
"""
File: middleware.py
Author: Rinat F Sabitov
Description: m3 license middleware
"""

from m3_license.api import get_project_lic_data


class LicenseMiddleware(object):
    """Предназначение простое: докидывать в объект
    реквеста данные из лицензии.
    Ожидается, что в settings уже прописаны LIC_KEY_FILE и PUBLIC_KEY
    """

    def process_request(self, request):
        """Добавляем в request содержимое лицензии
        """
        request.license = get_project_lic_data()



# coding: utf-8
"""
File: test_middleware.py
Author: Rinat F Sabitov
Description:
"""

import os
import tempfile
import datetime
from mock import Mock
from m3_license.middleware import LicenseMiddleware
from m3_license.api import LicenseException

from unittest import TestCase
import django.conf

from .env import LICENSE, PUBLIC_KEY


class LicenseMiddlewareTest(TestCase):

    def setUp(self):
        self.lm = LicenseMiddleware()
        self.request = Mock()
        tmp_file, self.tmp_filename = tempfile.mkstemp()
        with os.fdopen(tmp_file, 'w') as f:
            f.write(LICENSE)
        if not django.conf.settings.configured:
            django.conf.settings.configure(
                default_settings=django.conf.global_settings,
            )

    def test_process_request(self):
        django.conf.settings.LIC_KEY_FILE = self.tmp_filename
        django.conf.settings.PUBLIC_KEY = PUBLIC_KEY
        self.lm.process_request(self.request)
        self.assertEqual(self.request.license['organization'],
            u'Ключ для Некоммерческого использования Ветеринария')

        self.assertIsInstance(self.request.license['startdate'], datetime.date)
        self.assertIsInstance(self.request.license['enddate'], datetime.date)

    def test_exceptions(self):
        del django.conf.settings.LIC_KEY_FILE
        del django.conf.settings.PUBLIC_KEY

        self.assertRaises(LicenseException,
            self.lm.process_request, self.request)
        django.conf.settings.LIC_KEY_FILE = self.tmp_filename
        self.assertRaises(LicenseException,
            self.lm.process_request, self.request)
        del django.conf.settings.LIC_KEY_FILE
        django.conf.settings.PUBLIC_KEY = PUBLIC_KEY
        self.assertRaises(LicenseException,
            self.lm.process_request, self.request)



# coding: utf-8
"""
File: test_manage_command.py
Author: Rinat F Sabitov
Description:
"""


import os
import tempfile
from unittest import TestCase
import django.conf

from .env import LICENSE, PUBLIC_KEY, INVALID_SIGNATURE_LICENSE


class LicenseMiddlewareTest(TestCase):

    def setUp(self):
        if not django.conf.settings.configured:
            django.conf.settings.configure(
                default_settings=django.conf.global_settings,
            )

    def test_valid_license(self):
        from django.core.management import call_command
        tmp_file, tmp_filename = tempfile.mkstemp()
        with os.fdopen(tmp_file, 'w') as f:
            f.write(LICENSE)
        django.conf.settings.LIC_KEY_FILE = tmp_filename
        django.conf.settings.PUBLIC_KEY = PUBLIC_KEY
        django.conf.settings.INSTALLED_APPS = ['m3_license', ]

        with self.assertRaises(SystemExit) as cm:
            call_command('license_status')

        self.assertEqual(cm.exception.code, 0)

    def test_invalid_license(self):
        from django.core.management import call_command
        tmp_file, tmp_filename = tempfile.mkstemp()
        with os.fdopen(tmp_file, 'w') as f:
            f.write(INVALID_SIGNATURE_LICENSE)
        django.conf.settings.LIC_KEY_FILE = tmp_filename
        django.conf.settings.PUBLIC_KEY = PUBLIC_KEY
        django.conf.settings.INSTALLED_APPS = ['m3_license', ]

        with self.assertRaises(SystemExit) as cm:
            call_command('license_status')

        self.assertEqual(cm.exception.code, 1)



'''
File: __init__.py
Author: Rinat F Sabitov
Description: m3 license managements module
'''

import os
import subprocess

VERSION = (0, 0, 2)


def get_version(version=None):
    "Returns a PEP 386-compliant version number from VERSION."
    if version is None:
        version = VERSION
    else:
        assert len(version) == 5
        assert version[3] in ('alpha', 'beta', 'rc', 'final')

    parts = 2 if version[2] == 0 else 3
    main = '.'.join(str(x) for x in version[:parts])

    return str(main)


def get_hg_changeset():
    """Returns a numeric identifier of the latest hg changeset."""
    repo_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    hg_log = subprocess.Popen(
        'hg tip --template "{latesttagdistance}-{node|short}"',
        stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        shell=True, cwd=repo_dir, universal_newlines=True
    )
    return hg_log.communicate()[0]



# -*- coding: utf-8 -*-
"""
File: setup.py
Author: Rinat F Sabitov
Description: m3 license management module setup file

Tips:
 http://packages.python.org/distribute/setuptools.html
 http://diveintopython3.org/packaging.html
 http://wiki.python.org/moin/CheeseShopTutorial
 http://pypi.python.org/pypi?:action=list_classifiers
"""

import sys
import os
from setuptools import setup, find_packages

root_dir = os.path.dirname(__file__)
sys.path.insert(0, os.path.join(root_dir, 'src'))

version = __import__('m3_license').get_version()


def read(fname):
    """ Attempt to read a file and return it's content as a text
    """
    try:
        return open(os.path.join(os.path.dirname(__file__), fname)).read()
    except IOError:
        return ''

setup(
    name="m3-license",
    version=version,
    description=read('DESCRIPTION'),
    keywords="django m3 bars group",

    author="BARS Group",
    author_email="bars@bars-open.ru",

    maintainer='BARS Group',
    maintainer_email='bars@bars-open.ru',

    url="http://bars-open.ru",
    package_dir={'': 'src'},
    packages=find_packages("src"),
    classifiers=[
        'Intended Audience :: Developers',
        'Environment :: Web Environment',
        'Natural Language :: Russian',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'License :: OSI Approved :: MIT License',
        'Development Status :: 5 - Production/Stable',
    ],
    install_requires=read('REQUIREMENTS'),
    include_package_data=True,
    zip_safe=False,
    long_description=read('README'),
    test_suite='test',
    tests_request=['nose', ]
)



# coding: utf-8
"""
File: test_license.py
Author: Rinat F Sabitov
Description:
"""

import os
import unittest
from m3_license import api
import tempfile

from .env import LICENSE, PUBLIC_KEY

class GetDataTest(unittest.TestCase):

    def setUp(self):
        tmp_file, self.tmp_filename = tempfile.mkstemp()
        with os.fdopen(tmp_file, 'w') as f:
            f.write(LICENSE)


    def test_read_write(self):
        api.get_lic_data(self.tmp_filename, PUBLIC_KEY)



# -*- coding: utf-8 -*-
"""
Работа с файлами лицензий
"""


class LicenseException(Exception):
    pass


def get_project_lic_data():
    from django.conf import settings

    try:
        lic_filename = settings.LIC_KEY_FILE
    except AttributeError:
        raise LicenseException('`LIC_KEY_FILE` not defined in django '
            'settings file')

    try:
        public_key = settings.PUBLIC_KEY
    except AttributeError:
        raise LicenseException('`PUBLIC_KEY` not defined in django '
            'settings file')

    return get_lic_data(lic_filename, public_key)


def get_lic_data(xml_file, public_key):
    '''
    Функция читает файл лицензии, и если он корректен,
    возвращает словарь, содержащий данные о лицензии.
    Если файл лицензии некорректен - возвращается None
    @xml_file - имя файла лицензии
    @public_key - публичный ключ
    '''

    import os
    import rsa
    import re
    import copy
    import hashlib
    import htmlentitydefs
    import datetime
    import binascii
    from lxml import etree
    from rsa import pkcs1, transform
    from rsa.pkcs1 import VerificationError

    public_key = rsa.key.PublicKey.load_pkcs1_openssl_pem(public_key)

    NS = {
        'ds': 'http://www.w3.org/2000/09/xmldsig#'
    }
    TRANSFORM_ENVELOPED_SIGNATURE = (
        'http://www.w3.org/2000/09/xmldsig#enveloped-signature'
    )
    TRANSFORM_C14N_EXCLUSIVE_WITH_COMMENTS = (
        'http://www.w3.org/2001/10/xml-exc-c14n#WithComments'
    )
    TRANSFORM_C14N_EXCLUSIVE = 'http://www.w3.org/2001/10/xml-exc-c14n'
    TRANSFORM_C14N_INCLUSIVE = 'http://www.w3.org/TR/2001/REC-xml-c14n-20010315'

    _id_attributes = ['ID', 'id']

    def _get_by_id(t, id_v):
        for id_a in _id_attributes:
            elts = t.xpath("//*[@%s='%s']" % (id_a, id_v))
            if elts is not None and len(elts) > 0:
                return elts[0]
        return None

    def b64d(s):
        return s.decode('base64')

    def b64e(s):
        if type(s) in (int, long):
            s = transform.int2bytes(s)
        return s.encode('base64').replace('\n', '')

    def _alg(elt):
        uri = elt.get('Algorithm', None)
        if uri is None:
            return None
        else:
            return uri.rstrip('#')

    def _digest(str, hash_alg):
        h = getattr(hashlib, hash_alg)()
        h.update(str)
        digest = b64e(h.digest())
        return digest

    def _delete_elt(elt):
        if elt.getparent() is None:
            raise ValueError("Cannot delete root")
        if elt.tail is not None:
            p = elt.getprevious()
            if p is not None:
                if p.tail is None:
                    p.tail = ''
                p.tail += elt.tail
            else:
                up = elt.getparent()
                if up is None:
                    raise ValueError("Signature has no parent")
                if up.text is None:
                    up.text = ''
                up.text += elt.tail
        elt.getparent().remove(elt)

    def _remove_child_comments(t):
        root = t.getroot()
        for c in root.iter():
            if c.tag is etree.Comment or c.tag is etree.PI:
                _delete_elt(c)
        return t

    def _unescape(text):
        def fixup(m):
            text = m.group(0)
            if text[:2] == "&#":
                # character reference
                try:
                    if text[:3] == "&#x":
                        return unichr(int(text[3:-1], 16))
                    else:
                        return unichr(int(text[2:-1]))
                except ValueError:
                    pass
            else:
                # named entity
                try:
                    if text not in ('&amp;', '&lt;', '&gt;'):
                        text = unichr(htmlentitydefs.name2codepoint[text[1:-1]])
                except KeyError:
                    pass
            return text  # leave as is
        return re.sub("&#?\w+;", fixup, text)

    def _c14n(t, exclusive, with_comments, inclusive_prefix_list=None):
        cxml = etree.tostring(
            t, method="c14n",
            exclusive=exclusive,
            with_comments=with_comments
        )
        u = _unescape(cxml.decode("utf8", 'replace')).encode("utf8").strip()
        if u[0] != '<':
            raise ValueError("C14N buffer doesn't start with '<'")
        if u[-1] != '>':
            raise ValueError("C14N buffer doesn't end with '>'")
        return u

    def _enveloped_signature(t):
        sig = t.find(".//{%s}Signature" % NS['ds'])
        _delete_elt(sig)
        res = etree.tostring(t)
        u = _unescape(res.decode("utf8", 'replace')).encode("utf8").strip()
        return u

    def _transform(uri, t, tr=None):
        if uri == TRANSFORM_ENVELOPED_SIGNATURE:
            return _enveloped_signature(t)

        if uri in (
            TRANSFORM_C14N_EXCLUSIVE_WITH_COMMENTS,
            TRANSFORM_C14N_EXCLUSIVE
        ):
            nslist = None
            if tr is not None:
                elt = tr.find(
                    ".//{%s}InclusiveNamespaces"
                    % 'http://www.w3.org/2001/10/xml-exc-c14n#')
                if elt is not None:
                    nslist = elt.get('PrefixList', '').split()
            return _c14n(
                t, exclusive=True,
                with_comments=(uri == TRANSFORM_C14N_EXCLUSIVE_WITH_COMMENTS),
                inclusive_prefix_list=nslist
            )

        if uri == TRANSFORM_C14N_INCLUSIVE:
            return _c14n(t, exclusive=False, with_comments=False)

        raise ValueError("unknown or unimplemented transform %s" % uri)

    def _process_references(t, sig=None):
        if sig is None:
            sig = t.find(".//{%s}Signature" % NS['ds'])

        for ref in sig.findall(".//{%s}Reference" % NS['ds']):
            obj = None
            uri = ref.get('URI', None)
            if uri is None or uri == '#' or uri == '':
                ct = _remove_child_comments(copy.deepcopy(t))
                obj = ct.getroot()
            elif uri.startswith('#'):
                ct = copy.deepcopy(t)
                obj = _get_by_id(ct, uri[1:])
            else:
                raise ValueError("Unknown reference %s" % uri)

            if obj is None:
                raise ValueError(
                    "Unable to dereference Reference URI='%s'" % uri)

            for tr in ref.findall(".//{%s}Transform" % NS['ds']):
                obj = _transform(_alg(tr), obj, tr)

            dm = ref.find(".//{%s}DigestMethod" % NS['ds'])
            if dm is None:
                raise ValueError("Unable to find DigestMethod")
            hash_alg = _alg(dm).split("#")[1]
            digest = _digest(obj, hash_alg)
            dv = ref.find(".//{%s}DigestValue" % NS['ds'])
            dv.text = digest

    def check_lic_file(xml_file):
        try:
            doc = etree.parse(xml_file)
            sig = doc.getroot().find(".//{%s}Signature" % NS['ds'])
            if sig is None:
                raise ValueError("Is not signed xml!")
            else:
                _process_references(doc, sig)
                si = sig.find(".//{%s}SignedInfo" % NS['ds'])
                cm = si.find(".//{%s}CanonicalizationMethod" % NS['ds'])
                cm_alg = _alg(cm)
                if cm is None or cm_alg is None:
                    raise ValueError("No CanonicalizationMethod")
                sic = _transform(cm_alg, si)
                unsigned_xml = sic.replace(' xmlns=""', '')
                sve = sig.find(".//{%s}SignatureValue" % NS['ds'])
                signature = b64d(sve.text)
                return pkcs1.verify(unsigned_xml, signature, public_key)
        except (ValueError, binascii.Error):
            return False

    # ------------------------------------------------
    try:
        check_lic_file(xml_file)
    except VerificationError as err:
        raise LicenseException("License verification error: %s" % err)

    data = {}
    if not os.access(xml_file, os.R_OK):
        raise LicenseException("Read permission denied: %s" % xml_file)
    doc = etree.parse(xml_file)
    if doc is None or doc.getroot() is None:
        if doc is not None:
            doc.freeDoc()
        return data
    element = doc.find('//LicenceData')
    for e in element:
        tag = e.tag

        if tag in ('startdate', 'enddate'):
            data[tag] = datetime.datetime.strptime(e.text, '%d.%m.%Y')
        else:
            data[tag] = e.text
    return data



# coding: utf-8
"""
File: license_status.py
Author: Rinat F Sabitov
Description:
"""

from django.core.management.base import BaseCommand
import sys


class Command(BaseCommand):
    """Простая проверка лицензии на валидность.
    Исходя из ее результатов будут приниматься
    решения об обновлении системы
    """

    def handle(self, *args, **options):
        from m3_license.api import get_project_lic_data, LicenseException
        try:
            get_project_lic_data()
        except LicenseException as err:
            sys.stderr.write(unicode(err))
            sys.exit(1)
        sys.exit(0)
